%%% @doc
%%% GajuMine Setup GUI
%%%
%%% In the common case, when a user first runs the program they will not have any
%%% accounrts, keys, or mining system setup. The purpose of this program is to provide
%%% a first-run guide for them to get set up for mining and explain to the user what is
%%% going on.
%%%
%%% There are two cases to cover:
%%% 1. A new miner who doesn't know how anything works and needs to get set up from scratch.
%%% 2. A miner who does know a bit about how things work and wants to import a key.
%%% @end

-module(gmc_setup).
-vsn("0.1.0").
-author("Craig Everett <craigeverett@qpq.swiss>").
-copyright("Craig Everett <craigeverett@qpq.swiss>").
-license("GPL-3.0-or-later").

-behavior(wx_object).
-include_lib("wx/include/wx.hrl").
-export([trouble/1, done/0, done/1]).
-export([start_link/1]).
-export([init/1, terminate/2, code_change/3,
         handle_call/3, handle_cast/2, handle_info/2, handle_event/2]).
-include("$zx_include/zx_logger.hrl").


-record(w,
        {name = none :: atom(),
         id   = 0    :: integer(),
         wx   = none :: none | wx:wx_object()}).

-record(s,
        {wx      = none :: none | wx:wx_object(),
         frame   = none :: none | wx:wx_object(),
         lang    = en   :: en | jp,
         j       = none :: none | fun(),
         buttons = []   :: [#w{}]}).



%%% Interface functions

trouble(Info) ->
    wx_object:cast(?MODULE, {trouble, Info}).


done() ->
    wx_object:cast(?MODULE, done).


done(Mnemonic) ->
    wx_object:cast(?MODULE, {done, Mnemonic}).



%%% Startup Functions

start_link(Prefs) ->
    wx_object:start_link({local, ?MODULE}, ?MODULE, Prefs, []).


init(Prefs) ->
    ok = log(info, "GUI starting..."),
    Lang = maps:get(lang, Prefs, en_US),
    Trans = gd_jt:read_translations(?MODULE),
    J = gd_jt:j(Lang, Trans),

    WX = wx:new(),
    Frame = wxFrame:new(WX, ?wxID_ANY, J("GajuMine Setup")),
    MainSz = wxBoxSizer:new(?wxVERTICAL),

    ButtonTemplates =
        [{new_key,  J("Create a New Account")},
         {recover,  J("Recover Existing Account")}],

    MakeButton =
        fun({Name, Label}) ->
            B = wxButton:new(Frame, ?wxID_ANY, [{label, Label}]),
            #w{name = Name, id = wxButton:getId(B), wx = B}
        end,

    Buttons = lists:map(MakeButton, ButtonTemplates),

    Add = fun(#w{wx = Button}) -> wxBoxSizer:add(MainSz, Button, zxw:flags(wide)) end,
    ok = lists:foreach(Add, Buttons),

    ok = wxFrame:setSizer(Frame, MainSz),
    ok = wxSizer:layout(MainSz),

    ok = wxFrame:connect(Frame, command_button_clicked),
    ok = wxFrame:connect(Frame, close_window),
    ok = wxFrame:setSize(Frame, {300, 300}),
    ok = wxFrame:center(Frame),
    true = wxFrame:show(Frame),
    State =
        #s{wx      = WX,     frame = Frame,
           lang    = Lang,   j     = J,
           buttons = Buttons},
    {Frame, State}.


handle_call(Unexpected, From, State) ->
    ok = log(warning, "Unexpected call from ~tp: ~tp~n", [From, Unexpected]),
    {noreply, State}.


handle_cast({trouble, Info}, State) ->
    ok = handle_troubling(State, Info),
    {noreply, State};
handle_cast(done, State = #s{frame = Frame}) ->
    ok = gmc_con:end_setup(),
    ok = wxWindow:destroy(Frame),
    {noreply, State};
handle_cast({done, Mnemonic}, State) ->
    ok = done(Mnemonic, State),
    {noreply, State};
handle_cast(Unexpected, State) ->
    ok = log(warning, "Unexpected cast: ~tp~n", [Unexpected]),
    {noreply, State}.


handle_info(Unexpected, State) ->
    ok = log(warning, "Unexpected info: ~tp~n", [Unexpected]),
    {noreply, State}.


handle_event(#wx{event = #wxCommand{type = command_button_clicked},
                 id = ID},
             State = #s{buttons = Buttons}) ->
    ok =
        case lists:keyfind(ID, #w.id, Buttons) of
            #w{name = new_key} -> new_key(State);
            #w{name = recover} -> recover(State)
        end,
    {noreply, State};
handle_event(#wx{event = #wxClose{}}, State = #s{frame = Frame}) ->
    ok = gmc_con:stop(),
    ok = wxWindow:destroy(Frame),
    {noreply, State};
handle_event(Event, State) ->
    ok = tell(info, "Unexpected event ~tp State: ~tp~n", [Event, State]),
    {noreply, State}.


handle_troubling(#s{frame = Frame}, Info) ->
    zxw:show_message(Frame, Info).


code_change(_, State, _) ->
    {ok, State}.


terminate(Reason, State) ->
    ok = log(info, "Reason: ~tp, State: ~tp", [Reason, State]),
    wx:destroy().


new_key(State = #s{frame = Frame, j = J}) ->
    case new_key(Frame, J) of
        {ok, Phrase, Network} ->
            gmc_con:make_key(Phrase, Network);
        mismatch ->
            ok = zxw:show_message(Frame, J("Password entered incorrectly. Try again.")),
            new_key(State);
        cancel ->
            ok
    end.


new_key(Frame, J) ->
    Label = J("Generate New Account Key"),
    Dialog = wxDialog:new(Frame, ?wxID_ANY, Label, [{size, {500, 180}}]),
    Sizer = wxBoxSizer:new(?wxVERTICAL),
    Network = wxRadioBox:new(Dialog, ?wxID_ANY, J("Network"), {0, 0}, {50, 50}, [J("Mainnet"), J("Testnet")]),
    PassSz1 = wxStaticBoxSizer:new(?wxVERTICAL, Dialog, [{label, J("Create Password")}]),
    PassTx1 = wxTextCtrl:new(Dialog, ?wxID_ANY, [{style, ?wxTE_PASSWORD}]),
    _ = wxStaticBoxSizer:add(PassSz1, PassTx1, zxw:flags(wide)),
    PassSz2 = wxStaticBoxSizer:new(?wxVERTICAL, Dialog, [{label, J("Confirm Password")}]),
    PassTx2 = wxTextCtrl:new(Dialog, ?wxID_ANY, [{style, ?wxTE_PASSWORD}]),
    _ = wxStaticBoxSizer:add(PassSz2, PassTx2, zxw:flags(wide)),
    ButtSz = wxBoxSizer:new(?wxHORIZONTAL),
    Affirm = wxButton:new(Dialog, ?wxID_OK),
    Cancel = wxButton:new(Dialog, ?wxID_CANCEL),
    _ = wxBoxSizer:add(ButtSz, Affirm, zxw:flags(wide)),
    _ = wxBoxSizer:add(ButtSz, Cancel, zxw:flags(wide)),
    _ = wxBoxSizer:add(Sizer, Network, zxw:flags(base)),
    _ = wxBoxSizer:add(Sizer, PassSz1, zxw:flags(base)),
    _ = wxBoxSizer:add(Sizer, PassSz2, zxw:flags(base)),
    _ = wxBoxSizer:add(Sizer, ButtSz, zxw:flags(wide)),
    ok = wxDialog:setSizer(Dialog, Sizer),
    ok = wxDialog:setSize(Dialog, {400, 250}),
    ok = wxBoxSizer:layout(Sizer),
    ok = wxFrame:center(Dialog),
    Outcome =
        case wxDialog:showModal(Dialog) of
            ?wxID_OK ->
                Net =
                    case wxRadioBox:getSelection(Network) of
                        0 -> mainnet;
                        1 -> testnet;
                        _ -> mainnet
                    end,
                Pass1 = wxTextCtrl:getValue(PassTx1),
                Pass2 = wxTextCtrl:getValue(PassTx2),
                case Pass1 =:= Pass2 of
                    true  -> {ok, Pass1, Net};
                    false -> mismatch
                end;
            ?wxID_CANCEL ->
                cancel
        end,
    ok = wxDialog:destroy(Dialog),
    Outcome.


recover(State = #s{frame = Frame, j = J}) ->
    case recover(Frame, J) of
        {ok, Phrase, Mnemonic, Network} ->
            gmc_con:load_key(Phrase, Mnemonic, Network);
        mismatch ->
            ok = zxw:show_message(Frame, J("Password entered incorrectly. Try again.")),
            recover(State);
        cancel ->
            ok
    end.

recover(Frame, J) ->
    Dialog = wxDialog:new(Frame, ?wxID_ANY, J("Import Account Key")),
    Sizer = wxBoxSizer:new(?wxVERTICAL),
    Network = wxRadioBox:new(Dialog, ?wxID_ANY, J("Network"), {0, 0}, {50, 50}, [J("Mainnet"), J("Testnet")]),
    PassSz1 = wxStaticBoxSizer:new(?wxVERTICAL, Dialog, [{label, J("Create Password")}]),
    PassTx1 = wxTextCtrl:new(Dialog, ?wxID_ANY, [{style, ?wxTE_PASSWORD}]),
    _ = wxStaticBoxSizer:add(PassSz1, PassTx1, zxw:flags(wide)),
    PassSz2 = wxStaticBoxSizer:new(?wxVERTICAL, Dialog, [{label, J("Confirm Password")}]),
    PassTx2 = wxTextCtrl:new(Dialog, ?wxID_ANY, [{style, ?wxTE_PASSWORD}]),
    _ = wxStaticBoxSizer:add(PassSz2, PassTx2, zxw:flags(wide)),
    MnemSz = wxStaticBoxSizer:new(?wxVERTICAL, Dialog, [{label, J("Recovery Phrase")}]),
    MnemTx = wxTextCtrl:new(Dialog, ?wxID_ANY, [{style, ?wxTE_MULTILINE}]),
    _ = wxStaticBoxSizer:add(MnemSz, MnemTx, zxw:flags(wide)),
    ButtSz = wxBoxSizer:new(?wxHORIZONTAL),
    Affirm = wxButton:new(Dialog, ?wxID_OK),
    Cancel = wxButton:new(Dialog, ?wxID_CANCEL),
    _ = wxBoxSizer:add(ButtSz, Affirm, zxw:flags(wide)),
    _ = wxBoxSizer:add(ButtSz, Cancel, zxw:flags(wide)),
    _ = wxBoxSizer:add(Sizer, Network, zxw:flags(base)),
    _ = wxBoxSizer:add(Sizer, PassSz1, zxw:flags(base)),
    _ = wxBoxSizer:add(Sizer, PassSz2, zxw:flags(base)),
    _ = wxBoxSizer:add(Sizer, MnemSz, zxw:flags(wide)),
    _ = wxBoxSizer:add(Sizer, ButtSz, zxw:flags(base)),
    ok = wxDialog:setSizer(Dialog, Sizer),
    ok = wxDialog:setSize(Dialog, {400, 400}),
    ok = wxBoxSizer:layout(Sizer),
    ok = wxFrame:center(Dialog),
    Outcome =
        case wxDialog:showModal(Dialog) of
            ?wxID_OK ->
                Pass1 = wxTextCtrl:getValue(PassTx1),
                Pass2 = wxTextCtrl:getValue(PassTx2),
                case Pass1 =:= Pass2 of
                    true ->
                        Net =
                            case wxRadioBox:getSelection(Network) of
                                0 -> mainnet;
                                1 -> testnet;
                                _ -> mainnet
                            end,
                        Mnemonic = wxTextCtrl:getValue(MnemTx),
                        {ok, Pass1, Mnemonic, Net};
                    false ->
                        mismatch
                end;
            ?wxID_CANCEL ->
                cancel
        end,
    ok = wxDialog:destroy(Dialog),
    Outcome.



done(Mnemonic, #s{frame = Frame, j = J}) ->
    Dialog = wxDialog:new(Frame, ?wxID_ANY, J("Mnemonic")),
    Sizer = wxBoxSizer:new(?wxVERTICAL),
    MnemSz = wxStaticBoxSizer:new(?wxVERTICAL, Dialog, [{label, J("Recovery Phrase")}]),
    Options = [{value, Mnemonic}, {style, ?wxTE_MULTILINE bor ?wxTE_READONLY}],
    MnemTx = wxTextCtrl:new(Dialog, ?wxID_ANY, Options),
    _ = wxStaticBoxSizer:add(MnemSz, MnemTx, zxw:flags(wide)),
    ButtSz = wxBoxSizer:new(?wxHORIZONTAL),
    CloseB = wxButton:new(Dialog, ?wxID_CANCEL, [{label, J("Close")}]),
    CopyB  = wxButton:new(Dialog, ?wxID_OK, [{label, J("Copy to Clipboard")}]),
    _ = wxBoxSizer:add(ButtSz, CloseB, zxw:flags(wide)),
    _ = wxBoxSizer:add(ButtSz, CopyB, zxw:flags(wide)),
    _ = wxBoxSizer:add(Sizer, MnemSz, zxw:flags(wide)),
    _ = wxBoxSizer:add(Sizer, ButtSz, zxw:flags(base)),
    ok = wxDialog:setSizer(Dialog, Sizer),
    ok = wxBoxSizer:layout(Sizer),
    ok = wxFrame:center(Dialog),
    ok = wxStyledTextCtrl:setFocus(MnemTx),
    ok =
        case wxDialog:showModal(Dialog) of
            ?wxID_CANCEL -> ok;
            ?wxID_OK     -> copy_to_clipboard(Mnemonic)
        end,
    ok = wxDialog:destroy(Dialog),
    ok = gmc_con:end_setup(),
    ok = wxWindow:destroy(Frame),
    ok.


copy_to_clipboard(String) ->
    CB = wxClipboard:get(),
    case wxClipboard:open(CB) of
        true ->
            Text = wxTextDataObject:new([{text, String}]),
            case wxClipboard:setData(CB, Text) of
                true  ->
                    R = wxClipboard:flush(CB),
                    log(info, "String copied to system clipboard. Flushed: ~p", [R]);
                false ->
                    log(info, "Failed to copy to clipboard")
            end,
            ok = wxClipboard:close(CB);
        false ->
            log(info, "Failed to acquire the clipboard.")
    end.

